/*
    WMon
    Copyright 2004 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include <gtk/gtk.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "eggtrayicon.h"
#include "monitor.h"
#include "timer.h"

// TODO: try to find docker if it disappears

#define BROWSERCMD "mozilla -remote openURL\\(\"%s\",new-tab\\)"

#define ICON_DEFAULT GTK_STOCK_NETWORK
//#define ICON_DEFAULT GTK_STOCK_NEW
//#define ICON_UPDATED GTK_STOCK_DIALOG_WARNING
//#define ICON_UPDATED GTK_STOCK_DIALOG_INFO
#define ICON_UPDATED GTK_STOCK_NEW

static EggTrayIcon *icon;
static GtkWidget *image;
static vector<URL> list;
static bool modified;

static void iconDestroyed(GtkWidget *widget, void *data) {
	g_object_unref(G_OBJECT(icon));
	icon = NULL;
}

static void setIcon(const gchar *name) {
	gtk_image_set_from_stock(GTK_IMAGE(image), name, GTK_ICON_SIZE_LARGE_TOOLBAR);
	GdkPixbuf *pixbuf = gtk_widget_render_icon(image, name, GTK_ICON_SIZE_LARGE_TOOLBAR, NULL);
	GdkBitmap *mask;
	gdk_pixbuf_render_pixmap_and_mask(pixbuf,NULL,&mask,128);
	gtk_widget_shape_combine_mask(GTK_WIDGET(icon), mask, 0, 0 );
	g_object_unref(G_OBJECT(pixbuf));
	g_object_unref(G_OBJECT(mask));
}


static void updateMenu();

static void update() {
	list.clear();
	checkUpdates(list);
	int n = list.size();
	modified = false;
	for (int i = 0; i < n; i++) {
		if (list[i].isModified()) modified = true;
	}
	setIcon(modified?ICON_UPDATED:ICON_DEFAULT);
	updateMenu();
	Timer::waitStart();
}

static void clear() {
	clearUpdates();
	int n = list.size();
	for (int i = 0; i < n; i++) {
		list[i].setModified(false);
	}
	modified = false;
	setIcon(ICON_DEFAULT);
	updateMenu();
}

static GtkWidget *menu = NULL;
static void appendToMenu(GtkWidget *menu, const char *text, const char *icon, GtkSignalFunc callback, gpointer data = NULL) {
	GtkWidget *item;
	if (icon!=NULL) {
		item = gtk_image_menu_item_new_with_label(text);
		gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(item), gtk_image_new_from_stock(icon, GTK_ICON_SIZE_MENU));
	} else {
		item = gtk_menu_item_new_with_label(text);
	}
	if (callback!=NULL)
		g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(callback), data);
	gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
}

static void refreshClicked(GtkMenuItem *item, gpointer data) {
	Timer::runNow();
}

static void clearClicked(GtkMenuItem *item, gpointer data) {
	clear();
}

static void closeWindow(GtkButton *widget, gpointer data) {
	gtk_widget_destroy(GTK_WIDGET(data));
}

static void clearButton(GtkButton *widget, gpointer data) {
	gtk_widget_destroy(GTK_WIDGET(data));
	clear();
}

static void viewClicked(GtkMenuItem *item, gpointer data) {
	int n = list.size();
	int last = -1;
	for (int i = 0; i < n; i++) {
		if (list[i].isModified()) last = i;
	}

	GtkTextBuffer *textbuffer = gtk_text_buffer_new(NULL);
	for (int i = 0; i < n; i++) {
		if (!list[i].isModified()) continue;
		char buffer[200];
		gchar *utf8;
		if (list[i].isBinary()) {
			snprintf(buffer,sizeof(buffer),"%s has changed.\n",list[i].getURL());
			utf8 = g_locale_to_utf8(buffer, -1, NULL, NULL, NULL);
			gtk_text_buffer_insert_at_cursor(textbuffer, utf8, -1);
			g_free(utf8);
		} else {
			snprintf(buffer,sizeof(buffer),"%s has changed:\n",list[i].getURL());
			gchar *utf8 = g_locale_to_utf8(buffer, -1, NULL, NULL, NULL);
			gtk_text_buffer_insert_at_cursor(textbuffer, utf8, -1);
			g_free(utf8);
			utf8 = g_locale_to_utf8(list[i].getChanges(), -1, NULL, NULL, NULL);
			gtk_text_buffer_insert_at_cursor(textbuffer, utf8, -1);
			g_free(utf8);
		}
		if (i<last) {
			snprintf(buffer,sizeof(buffer),"\n---\n\n");
			utf8 = g_locale_to_utf8(buffer, -1, NULL, NULL, NULL);
			gtk_text_buffer_insert_at_cursor(textbuffer, utf8, -1);
			g_free(utf8);
		}
	}
	GtkWidget *textview = gtk_text_view_new_with_buffer(textbuffer);
	g_object_unref(G_OBJECT(textbuffer));
	gtk_text_view_set_editable(GTK_TEXT_VIEW(textview), FALSE);

	GtkWidget *window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	GtkWidget *box = gtk_vbox_new(FALSE, 10);
	GtkWidget *bbox = gtk_hbutton_box_new();

	GtkWidget *ok = gtk_button_new_from_stock(GTK_STOCK_OK);
	g_signal_connect(G_OBJECT(ok), "clicked", G_CALLBACK(closeWindow), (gpointer)window);
	GtkWidget *clear = gtk_button_new_from_stock(GTK_STOCK_CLEAR);
	g_signal_connect(G_OBJECT(clear), "clicked", G_CALLBACK(clearButton), (gpointer)window);

	gtk_container_add(GTK_CONTAINER(bbox), ok);
	gtk_container_add(GTK_CONTAINER(bbox), clear);

	GtkWidget *scrollwindow = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrollwindow), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrollwindow), textview);

	gtk_container_set_border_width(GTK_CONTAINER(box), 10);
	gtk_box_pack_start(GTK_BOX(box), scrollwindow, TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(box), bbox, FALSE, FALSE, 0);
	gtk_container_add(GTK_CONTAINER(window), box);
	gtk_window_set_default_size(GTK_WINDOW(window), 300, 350);
	gtk_window_set_title(GTK_WINDOW(window), "WMon - Changes");
	gtk_widget_show_all(window);
}

static void responseDialog(GtkDialog *dialog, gint arg1, gpointer data) {
	gtk_widget_destroy(GTK_WIDGET(dialog));
}

static GtkWidget *editText;

static void saveConfig(GtkButton *widget, gpointer data) {
	GtkTextBuffer *textbuffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(editText));

	char filename[500];
	getConfigFile(filename,sizeof(filename));
	FILE *f = fopen(filename,"w");
	if (f==NULL) {
		gchar *utf8 = g_locale_to_utf8(filename, -1, NULL, NULL, NULL);
		GtkWidget *dialog = gtk_message_dialog_new(NULL, (GtkDialogFlags)0, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, "Can't open %s", utf8);
		g_free(utf8);
		g_signal_connect(G_OBJECT(dialog), "response", G_CALLBACK(responseDialog), NULL);
		gtk_widget_show_all(dialog);
		return;
	}
	GtkTextIter start, end;
	gtk_text_buffer_get_start_iter(textbuffer, &start);
	gtk_text_buffer_get_end_iter(textbuffer, &end);
	gchar *text = gtk_text_buffer_get_text(textbuffer, &start, &end, FALSE);
	gchar *locale = g_locale_from_utf8(text, -1, NULL, NULL, NULL);
	fputs(locale, f);
	g_free(locale);
	g_free(text);
	fclose(f);


	gtk_widget_destroy(GTK_WIDGET(data));
	Timer::runNow();
}

static void editClicked(GtkMenuItem *item, gpointer data) {
	GtkTextBuffer *textbuffer = gtk_text_buffer_new(NULL);

	char filename[500];
	getConfigFile(filename,sizeof(filename));
	FILE *f = fopen(filename,"r");
	if (f==NULL) {
		gchar *utf8 = g_locale_to_utf8(filename, -1, NULL, NULL, NULL);
		GtkWidget *dialog = gtk_message_dialog_new(NULL, (GtkDialogFlags)0, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, "Can't open %s", utf8);
		g_free(utf8);
		g_signal_connect(G_OBJECT(dialog), "response", G_CALLBACK(responseDialog), NULL);
		gtk_widget_show_all(dialog);
		return;
	}
	while (true) {
		char buffer[500];
		if (fgets(buffer,sizeof(buffer),f)==NULL) break;
		gchar *utf8 = g_locale_to_utf8(buffer, -1, NULL, NULL, NULL);
		gtk_text_buffer_insert_at_cursor(textbuffer, utf8, -1);
		g_free(utf8);
	}
	fclose(f);


	editText = gtk_text_view_new_with_buffer(textbuffer);

	g_object_unref(G_OBJECT(textbuffer));
	gtk_text_view_set_editable(GTK_TEXT_VIEW(editText), TRUE);

	GtkWidget *window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	GtkWidget *box = gtk_vbox_new(FALSE, 10);
	GtkWidget *bbox = gtk_hbutton_box_new();

	GtkWidget *save = gtk_button_new_from_stock(GTK_STOCK_SAVE);
	g_signal_connect(G_OBJECT(save), "clicked", G_CALLBACK(saveConfig), (gpointer)window);
	GtkWidget *cancel = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
	g_signal_connect(G_OBJECT(cancel), "clicked", G_CALLBACK(closeWindow), (gpointer)window);

	gtk_container_add(GTK_CONTAINER(bbox), save);
	gtk_container_add(GTK_CONTAINER(bbox), cancel);

	GtkWidget *scrollwindow = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrollwindow), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrollwindow), editText);

	gtk_container_set_border_width(GTK_CONTAINER(box), 10);
	gtk_box_pack_start(GTK_BOX(box), scrollwindow, TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(box), bbox, FALSE, FALSE, 0);
	gtk_container_add(GTK_CONTAINER(window), box);
	gtk_window_set_default_size(GTK_WINDOW(window), 400, 250);
	gtk_window_set_title(GTK_WINDOW(window), "WMon - Config");
	gtk_widget_show_all(window);

}

static void quitClicked(GtkMenuItem *item, gpointer data) {
	exit(0);
}

static void urlClicked(GtkMenuItem *item, gpointer data) {
	int n = (int)data;
	char command[500];
	snprintf(command,sizeof(command),BROWSERCMD,list[n].getURL());
	system(command);
}

static void updateMenu() {
//	if (menu) g_object_unref(G_OBJECT(menu));
	menu = gtk_menu_new();
	int n = list.size();
	if (n>0)
		appendToMenu(menu, "Refresh", GTK_STOCK_REFRESH, G_CALLBACK(refreshClicked));
	if (modified) {
		appendToMenu(menu, "Clear updates", GTK_STOCK_CLEAR, G_CALLBACK(clearClicked));
		appendToMenu(menu, "View changes", GTK_STOCK_FIND, G_CALLBACK(viewClicked));
	}
	appendToMenu(menu, "Edit config", GTK_STOCK_PREFERENCES, G_CALLBACK(editClicked));

	if (n>0) {
		GtkWidget *submenu = gtk_menu_new();
		for (int i = 0; i < n; i++) {
			appendToMenu(submenu, list[i].getURL(), list[i].isModified()?GTK_STOCK_YES:NULL, G_CALLBACK(urlClicked), (gpointer)i);
		}
		GtkWidget *item = gtk_menu_item_new_with_label("URLs");
		gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), submenu);
		gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
	}

	gtk_menu_shell_append(GTK_MENU_SHELL(menu), gtk_separator_menu_item_new());
	appendToMenu(menu, "Quit", GTK_STOCK_QUIT, G_CALLBACK(quitClicked));
	gtk_widget_show_all(menu);
}

static void iconClicked(GtkWidget *widget, GdkEventButton *event, void *data) {
	gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL, event->button, gtk_get_current_event_time());
}


static gboolean timeoutFunc(gpointer data) {
	if (Timer::waitDone()) {
		update();
	}
	return TRUE;
}



int main(int argc, char *argv[]) {
	gtk_init(&argc, &argv);

	icon = egg_tray_icon_new("WMon");
	image = gtk_image_new();
	GtkWidget *box = gtk_event_box_new();
	gtk_container_add(GTK_CONTAINER(box),image);
	gtk_container_add(GTK_CONTAINER(icon),box);

	setIcon(ICON_DEFAULT);

	g_timeout_add(100,timeoutFunc,NULL);

	g_signal_connect(G_OBJECT(icon), "destroy", G_CALLBACK(iconDestroyed), NULL);
	g_signal_connect(G_OBJECT(box), "button-press-event", G_CALLBACK(iconClicked), NULL);


	gtk_widget_show_all(GTK_WIDGET(icon));
	Timer::runNow();
	gtk_main();

	return 0;
}

