/*
    WMon
    Copyright 2004 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <dirent.h>
#include <errno.h>
#include <stdlib.h>
#include <vector>
#include "monitor.h"
#include "timer.h"

using std::vector;

#define CONFIGFILENAME "config"

void checkDir(const char *dirname) {
	DIR *d = opendir(dirname);
	if (d!=NULL) {
		closedir(d);
		return;
	}
	if (errno==ENOENT) {
		if (mkdir(dirname,0700)==0) {
			char buffer[500];
			snprintf(buffer,sizeof(buffer),"%s/%s",dirname,CONFIGFILENAME);
			FILE *f = fopen(buffer,"w");
			if (f==NULL) {
				perror(buffer);
				exit(1);
			}
			fprintf(f,"interval = 900\n");
			fprintf(f,"#url = http://www.martin.st/\n");
			fprintf(f,"#binurl = http://foo/file.jpg\n");
			fclose(f);
			return;
		}
	}
	perror(dirname);
	exit(1);
}

void multArray(unsigned char *array, int size, int n) {
	int carry = 0;
	for (int i = size-1; i >= 0; i--) {
		carry += array[i]*n;
		array[i] = carry&0xFF;
		carry >>= 8;
	}
}

void addArray(unsigned char *array, int size, int n) {
	int carry = n;
	for (int i = size-1; i >= 0; i--) {
		carry += array[i];
		array[i] = carry&0xFF;
		carry >>= 8;
	}
}

void makeHash(const unsigned char *str, int len, unsigned char *hash, int size) {
	for (int pos = 0; pos < len || (len<0 && str[pos]!='\0'); pos++) {
/*		int shift = 5 + (str[pos]%5);
		unsigned int carry = hash[0]<<shift;
		carry >>= 8;
		for (int i = size-1; i >= 0; i--) {
			carry |= hash[i]<<shift;
			hash[i] = (carry)&0xFF;
			carry >>= 8;
		}
		hash[size-1] ^= str[pos];
*/
		multArray(hash,size,37);
		addArray(hash,size,(str[pos]+1));
	}
}

void makeHash(const char *str, unsigned char *hash, int size) {
	memset(hash,0,size);
	makeHash((const unsigned char*)str,-1,hash,size);
}

void hashStr(const unsigned char *hash, int size, char *str) {
	for (int i = 0; i < size; i++)
		sprintf(&str[2*i],"%02x",hash[i]);
}

URL::URL(const char *u, bool b) {
	url = strdup(u);
	makeHash(url,hash,16);
	hashStr(hash,16,hashText);
	modified = false;
	changes = NULL;
	binary = b;
}
URL::URL(const URL &u) {
	url = strdup(u.url);
	memcpy(hash,u.hash,sizeof(hash));
	memcpy(hashText,u.hashText,sizeof(hashText));
	modified = u.modified;
	if (u.changes==NULL) changes = NULL;
	else changes = strdup(u.changes);
	binary = u.binary;
}
URL::~URL() {
	free((void*)url);
	free((void*)changes);
}
const char *URL::getHashText() {
	return hashText;
}
const char *URL::getURL() {
	return url;
}
void URL::setChanges(const char *c) {
	if (changes!=NULL) free((void*)changes);
	changes = strdup(c);
}
void URL::setModified(bool m) {
	modified = m;
}
bool URL::isModified() {
	return modified;
}
const char *URL::getChanges() {
	return changes;
}
bool URL::isBinary() {
	return binary;
}



void readConfig(const char *filename, vector<URL> &list) {
	FILE *f = fopen(filename,"r");
	if (f==NULL) return;
	while (true) {
		char buffer[1000];
		if (fgets(buffer,sizeof(buffer),f)==NULL) break;
		char *ptr = buffer;
		if (*ptr=='#') continue;
		int len = strlen(ptr);
		if (len==0) continue;
		if (ptr[len-1]=='\n' || ptr[len-1]=='\r') ptr[--len] = '\0';
		if (ptr[len-1]=='\n' || ptr[len-1]=='\r') ptr[--len] = '\0';
		if (len==0) continue;

		while (*ptr==' ') ptr++;
		char *key = ptr;
		while (*ptr!=' ' && *ptr!='\0') ptr++;

		if (*ptr=='\0') continue;
		*ptr = '\0';
		ptr++;
		while (*ptr==' ') ptr++;
		if (*ptr!='=') continue;
		ptr++;
		while (*ptr==' ') ptr++;

		char *value = ptr;
		if (*value=='\0') continue;

		if (!strcmp(key,"interval")) {
			Timer::setInterval(atoi(ptr));
		} else if (!strcmp(key,"url")) {
			list.push_back(URL(ptr));
		} else if (!strcmp(key,"binurl")) {
			list.push_back(URL(ptr,true));
		}
	}
	fclose(f);
}

void cleanOld(const char *dirname, const char *config, vector<URL> &list) {
	DIR *d = opendir(dirname);
	if (d==NULL) {
		perror(dirname);
		exit(1);
	}
	int n = list.size();
	while (true) {
		struct dirent *entry = readdir(d);
		if (entry==NULL) break;
		if (!strcmp(entry->d_name,".")) continue;
		if (!strcmp(entry->d_name,"..")) continue;
		if (!strcmp(entry->d_name,config)) continue;
		int i = 0;
		for (i = 0; i < n; i++) {
			const char *filename = list[i].getHashText();
			if (!strcmp(entry->d_name,filename)) break;
			char buffer[500];
			snprintf(buffer,sizeof(buffer),"%s.new",filename);
			if (!strcmp(entry->d_name,buffer)) break;
		}
		if (i<n) continue;
		char name[500];
		snprintf(name,sizeof(name),"%s/%s",dirname,entry->d_name);
		unlink(name);
	}
	closedir(d);
}

bool fetch(const char *url, const char *filename) {
	pid_t pid;
	if ((pid = fork()) == 0) {
		execlp("wget","wget",url,"-O",filename,"-q",NULL);
		perror("wget");
		exit(42);
	}
	int status;
	while (true) {
		if (wait(&status)>=0) break;
		if (errno!=EINTR) break;
	}
	int retval = WEXITSTATUS(status);
	return retval==0;
}

void getMD5(const char *filename, char *md5) {
	char command[800];
	snprintf(command,sizeof(command),"md5sum %s",filename);
	FILE *in = popen(command,"r");
	if (in==NULL) {
		perror(filename);
		return;
	}
	char buffer[100];
	if (fgets(buffer,sizeof(buffer),in)==NULL) {
		fclose(in);
		return;
	}
	fclose(in);
	if (strlen(buffer)<32) {
		fclose(in);
		return;
	}
	strncpy(md5,buffer,32);
	buffer[32] = '\0';
}

void hashFile(const char *filename, char *hashText) {
	FILE *in = fopen(filename,"rb");
	if (in==NULL) {
		perror(filename);
		return;
	}
	unsigned char hash[16];
	memset(hash,0,sizeof(hash));
	while (true) {
		unsigned char buffer[4096];
		int n = fread(buffer,1,sizeof(buffer),in);
		if (n<=0) break;
		makeHash((const unsigned char*)buffer,n,hash,sizeof(hash));
	}
	fclose(in);
	hashStr(hash,16,hashText);
}

void makeText(const char *html, const char *txt) {
	char command[1500];
	snprintf(command,sizeof(command),"lynx -dump -force_html %s > %s",html,txt);
	system(command);
}

const char *diffFiles(const char *txt1, const char *txt2) {
	char command[1500];
	snprintf(command,sizeof(command),"diff -u %s %s | grep -v ^--- | grep -v ^+++",txt1,txt2);
	FILE *in = popen(command,"r");
	if (in==NULL) return NULL;
	int pos = 0;
	int size = 1000;
	char *buffer = (char*) malloc(size+1);
	while (true) {
		int n = fread(buffer+pos,1,size-pos,in);
		if (n<=0) break;
		pos += n;
		if (pos==size) {
			size *= 2;
			char *newbuffer = (char*) realloc(buffer,size+1);
			if (newbuffer==NULL) {
				free(buffer);
				fprintf(stderr,"Out of memory\n");
				return NULL;
			}
			buffer = newbuffer;
		}
	}
	buffer[pos] = '\0';
	fclose(in);
	return buffer;
}

void check(const char *dirname, URL &url) {
	char filename[500], newfile[500];
	snprintf(filename,sizeof(filename),"%s/%s",dirname,url.getHashText());
	snprintf(newfile,sizeof(newfile),"%s.new",filename);
	FILE *f = fopen(filename,"r");
	char origmd5[2*16+1] = "";
	if (f!=NULL) {
		getMD5(filename,origmd5);
		fclose(f);
	}

	if (!fetch(url.getURL(),newfile)) {
		FILE *f = fopen(newfile,"w");
		if (f!=NULL) fclose(f);
	}

	if (origmd5[0]=='\0') {
		rename(newfile,filename);
		return;
	}
	char newmd5[2*16+1] = "";
	getMD5(newfile,newmd5);
	if (newmd5[0]=='\0' || !strcmp(origmd5,newmd5)) {
		unlink(newfile);
		return;
	}

	url.setModified(true);
	if (url.isBinary()) return;

	char textfile1[500], textfile2[500];
	snprintf(textfile1,sizeof(textfile1),"%s.1.txt",filename);
	snprintf(textfile2,sizeof(textfile2),"%s.2.txt",filename);

	makeText(filename,textfile1);
	makeText(newfile,textfile2);

	const char *changes = diffFiles(textfile1,textfile2);
	url.setChanges(changes);
	free((void*)changes);

	unlink(textfile1);
	unlink(textfile2);
}

void initURLs(const char *dirname, vector<URL> &list) {
	checkDir(dirname);
	char config[500];
	snprintf(config,sizeof(config),"%s/%s",dirname,CONFIGFILENAME);
	readConfig(config,list);
	cleanOld(dirname,CONFIGFILENAME,list);
}

void getDir(char *dirname, int size) {
	const char *home = getenv("HOME");
	snprintf(dirname,size,"%s%s.wmon",home!=NULL?home:"",home!=NULL?"/":"");
}

void checkUpdates(vector<URL> &list) {
	char dirname[500];
	getDir(dirname,sizeof(dirname));
	initURLs(dirname,list);

	int n = list.size();
	for (int i = 0; i < n; i++) {
		check(dirname,list[i]);
	}

}


void clearUpdates() {
	char dirname[500];
	getDir(dirname,sizeof(dirname));
	vector<URL> list;
	initURLs(dirname,list);

	int n = list.size();
	for (int i = 0; i < n; i++) {
		URL &url = list[i];
		char filename[500], newfile[500];
		snprintf(filename,sizeof(filename),"%s/%s",dirname,url.getHashText());
		snprintf(newfile,sizeof(newfile),"%s.new",filename);
		FILE *f = fopen(newfile,"r");
		if (f==NULL) continue;
		fclose(f);
		rename(newfile,filename);
	}
}

void getConfigFile(char *buffer, int size) {
	char dirname[500];
	getDir(dirname,sizeof(dirname));
	snprintf(buffer,size,"%s/%s",dirname,CONFIGFILENAME);
}



