/*
    WMon
    Copyright 2004 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <windowsx.h>
#include <wininet.h>
#include <shellapi.h>
#include <stdio.h>
#include <stdlib.h>
#include "resource.h"
#include "timer.h"
#include "monitor.h"

#define SYSTRAY_ID 0
#define SYSTRAY_MSG WM_USER+SYSTRAY_ID

static HINSTANCE ghInst;
static HWND ghWnd;
static HWND ghEdit = NULL;

static vector<URL> list;
static bool modified;

void setIcon(int icon);

static void update() {
	list.clear();
	checkUpdates(list);
	int n = list.size();
	modified = false;
	for (int i = 0; i < n; i++) {
		if (list[i].isModified()) modified = true;
	}
	setIcon(modified?ICON_UPDATED:ICON_DEFAULT);
	Timer::waitStart();
}

static void clear() {
	clearUpdates();
	int n = list.size();
	for (int i = 0; i < n; i++) {
		list[i].setModified(false);
	}
	modified = false;
	setIcon(ICON_DEFAULT);
}

#define MENU_REFRESH 1
#define MENU_CLEAR 2
#define MENU_EDIT 3
#define MENU_EXIT 4
#define MENU_URL_START 20

bool downloadFile(const char *url, const char *filename) {
	HINTERNET agent = InternetOpen("WMon", INTERNET_OPEN_TYPE_PRECONFIG, NULL, NULL, 0);
	if (agent==NULL) return false;
	HINTERNET file = InternetOpenUrl(agent, url, NULL, (DWORD)-1, INTERNET_FLAG_NO_UI | INTERNET_FLAG_NO_COOKIES | INTERNET_FLAG_NO_CACHE_WRITE | INTERNET_FLAG_RELOAD | INTERNET_FLAG_RESYNCHRONIZE, 0);
	if (file==NULL) {
		InternetCloseHandle(agent);
		return false;
	}
	DWORD status, len;
	len = sizeof(status);
	HttpQueryInfo(file, HTTP_QUERY_STATUS_CODE | HTTP_QUERY_FLAG_NUMBER, &status, &len, NULL);
	if (status>=300) {
		InternetCloseHandle(file);
		InternetCloseHandle(agent);
		return false;
	}

	FILE *out = fopen(filename,"wb");
	if (out==NULL) {
		perror(filename);
		InternetCloseHandle(file);
		InternetCloseHandle(agent);
		return false;
	}
	while (true) {
		unsigned char buffer[2048];
		DWORD read;
		if (!InternetReadFile(file,buffer,sizeof(buffer),&read)) break;
		if (read<=0) break;
		fwrite(buffer,read,1,out);
	}
	fclose(out);
	InternetCloseHandle(file);
	InternetCloseHandle(agent);
	return true;
}

void setIcon(int icon) {
	NOTIFYICONDATA nid;
	nid.cbSize = sizeof(nid);
	nid.hWnd = ghWnd;
	nid.uID = SYSTRAY_ID;
	nid.uFlags = NIF_ICON;
	nid.hIcon = LoadIcon(ghInst, MAKEINTRESOURCE(icon));
	Shell_NotifyIcon(NIM_MODIFY, &nid);
	DestroyIcon(nid.hIcon);
}

void getConfig(HWND hCtrl) {
	char filename[500];
	getConfigFile(filename,sizeof(filename));
	FILE *f = fopen(filename,"rb");
	if (f==NULL) {
		char errbuffer[500];
		_snprintf(errbuffer,sizeof(errbuffer),"Can't open %s",filename);
		MessageBox(NULL, errbuffer, "Error", MB_OK | MB_ICONERROR);
		return;
	}
	int size = 5000;
	int pos = 0;
	char *buffer = (char*) malloc(size+1);
	while (true) {
		int n = fread(buffer,1,size-pos,f);
		if (n <= 0) break;
		pos += n;
		if (pos==size) {
			size *= 2;
			buffer = (char*) realloc(buffer,size+1);
		}
	}
	fclose(f);
	buffer[pos] = '\0';

	Edit_SetText(hCtrl,buffer);
	free(buffer);
}

void saveConfig(HWND hCtrl) {
	int len = Edit_GetTextLength(hCtrl);
	char *buffer = (char*) malloc(len+2);
	Edit_GetText(hCtrl,buffer,len+1);
	char filename[500];
	getConfigFile(filename,sizeof(filename));
	FILE *f = fopen(filename,"wb");
	if (f==NULL) {
		char errbuffer[500];
		_snprintf(errbuffer,sizeof(errbuffer),"Can't open %s",filename);
		MessageBox(NULL, errbuffer, "Error", MB_OK | MB_ICONERROR);
		return;
	}
	fwrite(buffer,len,1,f);
	fclose(f);
	free(buffer);
	Timer::runNow();
}

BOOL CALLBACK EditDialogProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	switch (uMsg) {
	case WM_INITDIALOG:
		ghEdit = hDlg;
		getConfig(GetDlgItem(hDlg, DIALOG_TEXT));
		return TRUE;
	case WM_SYSCOMMAND:
		if (wParam == SC_CLOSE) {
			EndDialog(hDlg, 0);
			return TRUE;
		}
	case WM_COMMAND:
		if (LOWORD(wParam) == DIALOG_CANCEL) {
			EndDialog(hDlg, 0);
			return TRUE;
		} else if (LOWORD(wParam) == DIALOG_SAVE) {
			saveConfig(GetDlgItem(hDlg, DIALOG_TEXT));
			EndDialog(hDlg, 0);
			return TRUE;
		}
		break;
	case WM_DESTROY:
		ghEdit = NULL;
		break;
	}
	fflush(stdout);
	return FALSE;
}

void showMenu() {
	HMENU menu = CreatePopupMenu();
	HMENU submenu = CreatePopupMenu();
	int n = list.size();
	if (n>0) {
		AppendMenu(menu, MF_ENABLED | MF_STRING, MENU_REFRESH, "&Refresh");

		for (int i = 0; i < n; i++) {
			AppendMenu(submenu, MF_ENABLED | MF_STRING | (list[i].isModified()?MF_CHECKED:0), MENU_URL_START+i, list[i].getURL());
		}
	}

	if (modified)
		AppendMenu(menu, MF_ENABLED | MF_STRING, MENU_CLEAR, "&Clear");
	AppendMenu(menu, (ghEdit!=NULL?MF_GRAYED:MF_ENABLED) | MF_STRING, MENU_EDIT, "&Edit config");
	if (n>0)
		AppendMenu(menu, MF_ENABLED | MF_POPUP, (int)submenu, "&Urls");
	AppendMenu(menu, MF_SEPARATOR, 0, NULL);
	AppendMenu(menu, MF_ENABLED | MF_STRING, MENU_EXIT, "E&xit");

	POINT point;
	GetCursorPos(&point);
	// TPM_CENTERALIGN, TPM_LEFTALIGN, TPM_RIGHTALIGN
	SetForegroundWindow(ghWnd); // said to be some ms kb bugfix
	int val = (int)TrackPopupMenu(menu, TPM_RETURNCMD | TPM_LEFTBUTTON, point.x, point.y, 0, ghWnd, NULL);
	PostMessage(ghWnd, 0, 0, 0); // this too
	switch (val) {
	case MENU_REFRESH:
		Timer::runNow();
		break;
	case MENU_CLEAR:
		clear();
		break;
	case MENU_EDIT:
		DialogBoxParam(ghInst, MAKEINTRESOURCE(DIALOG_EDIT), ghWnd, EditDialogProc, 0);
		break;
	case MENU_EXIT:
		DestroyWindow(ghWnd);
		break;
	default:
		if (val>=MENU_URL_START && val<MENU_URL_START+n) {
			ShellExecute(ghWnd, NULL, list[val-MENU_URL_START].getURL(), NULL, NULL, SW_SHOW);
		}
		break;
	}
	DestroyMenu(menu);
	DestroyMenu(submenu);
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	switch (uMsg) {
	case WM_CREATE:
		// CreateWindow
		break;
	case WM_TIMER:
		if (Timer::waitDone())
			update();
		break;
	case WM_DESTROY:
		if (ghEdit!=NULL)
			EndDialog(ghEdit, 0);
		NOTIFYICONDATA nid;
		nid.cbSize = sizeof(nid);
		nid.hWnd = hWnd;
		nid.uID = SYSTRAY_ID;
		Shell_NotifyIcon(NIM_DELETE, &nid);
//		PostQuitMessage(0);
		exit(0);
		break;

	case SYSTRAY_MSG:
		switch (lParam) {
		case WM_LBUTTONUP:
		case WM_RBUTTONUP:
			showMenu();
//		case WM_LBUTTONDBLCLK:
//		case WM_RBUTTONBLKCLK:
			break;
		}
		break;
	}
	return DefWindowProc(hWnd, uMsg, wParam, lParam);
}

int WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR lpCmdLine, int nCmdShow) {

	ghInst = hInst;

	const char *classname = "MyClass";

	WNDCLASSEX wc;

	wc.cbSize = sizeof(WNDCLASSEX);
	wc.style = 0;
	wc.lpfnWndProc = WndProc;
	wc.cbClsExtra = wc.cbWndExtra = 0;
	wc.hInstance = hInst;
	wc.hIcon = LoadIcon(hInst, MAKEINTRESOURCE(ICON_MAIN));
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = (HBRUSH)(COLOR_WINDOW);
	wc.lpszMenuName = NULL;
	wc.lpszClassName = classname;
	wc.hIconSm = (HICON) LoadImage(hInst, MAKEINTRESOURCE(ICON_MAIN), IMAGE_ICON, GetSystemMetrics(SM_CXSMICON), GetSystemMetrics(SM_CYSMICON), LR_DEFAULTCOLOR | LR_LOADTRANSPARENT);
	RegisterClassEx(&wc);

	ghWnd = CreateWindowEx(0, classname, "WMon", WS_POPUP, CW_USEDEFAULT, 0, CW_USEDEFAULT, 0, NULL, NULL, hInst, NULL);
	DestroyIcon(wc.hIcon);
	DestroyIcon(wc.hIconSm);
	DestroyCursor(wc.hCursor);


	NOTIFYICONDATA nid;
	nid.cbSize = sizeof(nid);
	nid.hWnd = ghWnd;
	nid.uID = SYSTRAY_ID;
	nid.uFlags = NIF_ICON | NIF_TIP | NIF_MESSAGE;
	nid.uCallbackMessage = SYSTRAY_MSG;
	nid.hIcon = LoadIcon(hInst, MAKEINTRESOURCE(ICON_DEFAULT));
//	nid.hIcon = (HICON) LoadImage(hInst, MAKEINTRESOURCE(ICON_DEFAULT), IMAGE_ICON, GetSystemMetrics(SM_CXICON), GetSystemMetrics(SM_CYICON), LR_DEFAULTCOLOR | LR_LOADTRANSPARENT);
	strcpy(nid.szTip,"WMon");

//	SetPriorityClass(GetCurrentProcess(), IDLE_PRIORITY_CLASS);

	Shell_NotifyIcon(NIM_ADD, &nid);
	DestroyIcon(nid.hIcon);

	SetTimer(ghWnd, 0, 500, NULL);


	Timer::runNow();

	MSG msg;
	while (GetMessage(&msg, NULL, 0, 0)>0) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

	return 0;
}

