/*
    WMon
    Copyright 2004 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include <stdio.h>
#include <stdlib.h>
#include <direct.h>
#define WIN32_MEAN_AND_LEAN
#include <windows.h>
#include <vector>
#include "monitor.h"
#include "timer.h"

using std::vector;

#define CONFIGFILENAME "config"


void PERROR(const char *fmt, ...) {
	char errbuffer[500];
	va_list list;
	va_start(list, fmt);
	_vsnprintf(errbuffer,sizeof(errbuffer),fmt,list);
	va_end(list);
	MessageBox(NULL, errbuffer, "Error", MB_OK | MB_ICONERROR);
}


void checkDir(const char *dirname) {
	WIN32_FIND_DATA info;
	HANDLE h = FindFirstFile(dirname,&info);
	bool dir = false;
	if (h!=INVALID_HANDLE_VALUE) {
		dir = info.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY;
		FindClose(h);
	}
	if (dir) return;
	if (_mkdir(dirname)!=0) {
		PERROR("Can't create %s",dirname);
		exit(1);
	}
	char buffer[500];
	_snprintf(buffer,sizeof(buffer),"%s\\%s",dirname,CONFIGFILENAME);
	FILE *f = fopen(buffer,"w");
	if (f==NULL) {
		PERROR("Can't create %s",buffer);
		exit(1);
	}
	fprintf(f,"interval = 900\r\n");
	fprintf(f,"#url = http://www.martin.st/\r\n");
	fclose(f);
	return;
}

void multArray(unsigned char *array, int size, int n) {
	int carry = 0;
	for (int i = size-1; i >= 0; i--) {
		carry += array[i]*n;
		array[i] = carry&0xFF;
		carry >>= 8;
	}
}

void addArray(unsigned char *array, int size, int n) {
	int carry = n;
	for (int i = size-1; i >= 0; i--) {
		carry += array[i];
		array[i] = carry&0xFF;
		carry >>= 8;
	}
}

void makeHash(const unsigned char *str, int len, unsigned char *hash, int size) {
	for (int pos = 0; pos < len || (len<0 && str[pos]!='\0'); pos++) {
/*		int shift = 5 + (str[pos]%5);
		unsigned int carry = hash[0]<<shift;
		carry >>= 8;
		for (int i = size-1; i >= 0; i--) {
			carry |= hash[i]<<shift;
			hash[i] = (carry)&0xFF;
			carry >>= 8;
		}
		hash[size-1] ^= str[pos];
*/
		multArray(hash,size,37);
		addArray(hash,size,(str[pos]+1));
	}
}

void makeHash(const char *str, unsigned char *hash, int size) {
	memset(hash,0,size);
	makeHash((const unsigned char*)str,-1,hash,size);
}

void hashStr(const unsigned char *hash, int size, char *str) {
	for (int i = 0; i < size; i++)
		sprintf(&str[2*i],"%02x",hash[i]);
}

URL::URL(const char *u, bool b) {
	url = strdup(u);
	makeHash(url,hash,16);
	hashStr(hash,16,hashText);
	modified = false;
	changes = NULL;
	binary = b;
}
URL::URL(const URL &u) {
	url = strdup(u.url);
	memcpy(hash,u.hash,sizeof(hash));
	memcpy(hashText,u.hashText,sizeof(hashText));
	modified = u.modified;
	if (u.changes==NULL) changes = NULL;
	else changes = strdup(u.changes);
	binary = u.binary;
}
URL::~URL() {
	free((void*)url);
	free((void*)changes);
}
const char *URL::getHashText() {
	return hashText;
}
const char *URL::getURL() {
	return url;
}
void URL::setChanges(const char *c) {
	if (changes!=NULL) free((void*)changes);
	changes = strdup(c);
}
void URL::setModified(bool m) {
	modified = m;
}
bool URL::isModified() {
	return modified;
}
const char *URL::getChanges() {
	return changes;
}
bool URL::isBinary() {
	return binary;
}



void readConfig(const char *filename, vector<URL> &list) {
	FILE *f = fopen(filename,"r");
	if (f==NULL) return;
	while (true) {
		char buffer[1000];
		if (fgets(buffer,sizeof(buffer),f)==NULL) break;
		char *ptr = buffer;
		if (*ptr=='#') continue;
		int len = strlen(ptr);
		if (len==0) continue;
		if (ptr[len-1]=='\n' || ptr[len-1]=='\r') ptr[--len] = '\0';
		if (ptr[len-1]=='\n' || ptr[len-1]=='\r') ptr[--len] = '\0';
		if (len==0) continue;

		while (*ptr==' ') ptr++;
		char *key = ptr;
		while (*ptr!=' ' && *ptr!='\0') ptr++;

		if (*ptr=='\0') continue;
		*ptr = '\0';
		ptr++;
		while (*ptr==' ') ptr++;
		if (*ptr!='=') continue;
		ptr++;
		while (*ptr==' ') ptr++;

		char *value = ptr;
		if (*value=='\0') continue;

		if (!_stricmp(key,"interval")) {
			Timer::setInterval(atoi(ptr));
		} else if (!_stricmp(key,"url")) {
			list.push_back(URL(ptr));
		} else if (!_stricmp(key,"binurl")) {
			list.push_back(URL(ptr,true));
		}
	}
	fclose(f);
}

void cleanOld(const char *dirname, const char *config, vector<URL> &list) {
	char dirbuffer[500];
	_snprintf(dirbuffer,sizeof(dirbuffer),"%s\\*",dirname);
	WIN32_FIND_DATA info;
	HANDLE h = FindFirstFile(dirbuffer,&info);
	if (h==INVALID_HANDLE_VALUE) {
		PERROR("Can't open %s",dirname);
		exit(1);
	}
	int n = list.size();
	while (true) {
		const char *name = info.cFileName;
		if (name==NULL) break;
		do { // cheap fake of goto
			if (!strcmp(name,".")) break;
			if (!strcmp(name,"..")) break;
			if (!_stricmp(name,config)) break;
			int i = 0;
			for (i = 0; i < n; i++) {
				const char *filename = list[i].getHashText();
				if (!_stricmp(name,filename)) break;
				char buffer[500];
				_snprintf(buffer,sizeof(buffer),"%s.new",filename);
				if (!_stricmp(name,buffer)) break;
			}
			if (i<n) break;
			char completename[500];
			_snprintf(completename,sizeof(completename),"%s\\%s",dirname,name);
			_unlink(completename);
		} while (false);

		if (!FindNextFile(h,&info)) {
			if (GetLastError()==ERROR_NO_MORE_FILES) {
				break;
			} else {
				PERROR("Error reading files in %s",dirname);
				exit(1);
			}
		}
	}
	FindClose(h);
}

void hashFile(const char *filename, char *hashText) {
	FILE *in = fopen(filename,"rb");
	if (in==NULL) {
		PERROR("Can't open %s",filename);
		return;
	}
	unsigned char hash[16];
	memset(hash,0,sizeof(hash));
	while (true) {
		unsigned char buffer[4096];
		int n = fread(buffer,1,sizeof(buffer),in);
		if (n<=0) break;
		makeHash((const unsigned char*)buffer,n,hash,sizeof(hash));
	}
	fclose(in);
	hashStr(hash,16,hashText);
}

bool downloadFile(const char *url, const char *filename);

void check(const char *dirname, URL &url) {
	char filename[500], newfile[500];
	_snprintf(filename,sizeof(filename),"%s/%s",dirname,url.getHashText());
	_snprintf(newfile,sizeof(newfile),"%s.new",filename);
	FILE *f = fopen(filename,"r");
	char origHash[2*16+1] = "";
	if (f!=NULL) {
		hashFile(filename,origHash);
		fclose(f);
	}

	if (!downloadFile(url.getURL(),newfile)) {
		FILE *f = fopen(newfile,"w");
		if (f!=NULL) fclose(f);
	}

	if (origHash[0]=='\0') {
		rename(newfile,filename);
		return;
	}
	char newHash[2*16+1] = "";
	hashFile(newfile,newHash);
	if (newHash[0]=='\0' || !_stricmp(origHash,newHash)) {
		_unlink(newfile);
		return;
	}

	url.setModified(true);
	if (url.isBinary()) return;

	// no text diffs on win32 yet
}

void initURLs(const char *dirname, vector<URL> &list) {
	checkDir(dirname);
	char config[500];
	_snprintf(config,sizeof(config),"%s\\%s",dirname,CONFIGFILENAME);
	readConfig(config,list);
	cleanOld(dirname,CONFIGFILENAME,list);
}

void getDir(char *dirname, int size) {
	_snprintf(dirname,size,"wmon-data");
}

void checkUpdates(vector<URL> &list) {
	char dirname[500];
	getDir(dirname,sizeof(dirname));
	initURLs(dirname,list);

	int n = list.size();
	for (int i = 0; i < n; i++) {
		check(dirname,list[i]);
	}

}


void clearUpdates() {
	char dirname[500];
	getDir(dirname,sizeof(dirname));
	vector<URL> list;
	initURLs(dirname,list);

	int n = list.size();
	for (int i = 0; i < n; i++) {
		URL &url = list[i];
		char filename[500], newfile[500];
		_snprintf(filename,sizeof(filename),"%s\\%s",dirname,url.getHashText());
		_snprintf(newfile,sizeof(newfile),"%s.new",filename);
		FILE *f = fopen(newfile,"r");
		if (f==NULL) continue;
		fclose(f);
		_unlink(filename);
		rename(newfile,filename);
	}
}

void getConfigFile(char *buffer, int size) {
	char dirname[500];
	getDir(dirname,sizeof(dirname));
	_snprintf(buffer,size,"%s\\%s",dirname,CONFIGFILENAME);
}



